/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* This file contains all code that manages units. */

#include "conq.h"

/* This is not a limit, just sets initial allocation of unit objects
   and how many additional to get if more are needed.  Should be tuned
   for the system. */

#ifndef INITMAXUNITS
#define INITMAXUNITS 200
#endif

/* The list of available units. */

Unit *freeunits;

/* A scratch global. */

Unit *tmpunit;

/* Buffer for descriptions of units. */

char shortbuf[BUFSIZE];         /* buffer for short unit description */

char actorstatebuf[BUFSIZE];

/* Total number of units in existence. */

int numunits = 0;
int nextid = 1;            /* next number to be used for ids */
int occdeath[MAXUTYPES];  /* buffer for remembering occupant death */

int recent_dead_flushed = TRUE; /* Have units died since dead unit lists made */

int totunits;
int totutypes[MAXUTYPES];

/* Init gets a first block of units, and sets up the "independent side" list,
   since newly created units will appear on it. */

void
init_units()
{
    extern Task *freetasks;

    allocate_unit_block();
    if (freetasks == NULL) init_tasks();  /* hack */
    init_side_unithead(indepside);
}

/* Grab another block of unit objects to work with. */

allocate_unit_block()
{
    int i;
    Unit *unitblock = (Unit *) xmalloc(INITMAXUNITS * sizeof(Unit));

    for (i = 0; i < INITMAXUNITS; ++i) {
	unitblock[i].id = -1;
#if (MAXMTYPES > 2)
	/* Null this out for now, we'll do the malloc at unit creation. */
	unitblock[i].supply = NULL;
#endif
	unitblock[i].tooling = NULL;
	unitblock[i].plan = NULL;
	unitblock[i].next = &unitblock[i+1];
    }
    unitblock[INITMAXUNITS-1].next = NULL;
    freeunits = unitblock;
    Dprintf("Allocated space for %d units.\n", INITMAXUNITS);
    return TRUE;
}

/* The primitive unit creator, called by regular creator and also used to
   make the dummy units that sit at the heads of the per-side unit lists. */

Unit *
create_bare_unit(type)
int type;
{
    Unit *newunit;

    /* If our free list is empty, go and get some more units. */
    if (freeunits == NULL) {
	allocate_unit_block();
    }
    /* Take the first unit off the free list. */
    newunit = freeunits;
    freeunits = freeunits->next;
    /* Give it a valid type... */
    newunit->type = type;
    /* ...but an invalid id. */
    newunit->id = -1;
    return newunit;
}

/* The regular unit creation routine.  All the slots should have something
   reasonable in them, since individual units objects see a lot of reuse. */

Unit *
create_unit(type)
int type;
{
    int i, r;
    Unit *newunit;

    /* Test whether we've hit any designer-specified limits. */
    if ((u_type_in_game_max(type) >= 0
	 && totutypes[type] >= u_type_in_game_max(type))
	|| (g_units_in_game_max() >= 0
	    && totunits >= g_units_in_game_max())) {
	return NULL;
    }
    /* Allocate the unit object.  Xconq will fail instead of returning null. */
    newunit = create_bare_unit(type);
    /* Init all the slots to distinguishable values. */
    newunit->name = NULL;
    newunit->x = newunit->y = -1;
    newunit->z = 0;  /* on the ground */
    /* Units default to being independent. */
    newunit->side = NULL;
    newunit->id = nextid++;
    /* Number == 0 means unit is unnumbered. */
    newunit->number = 0;
    /* Create at max hp (caller should reduce if necessary). */
    newunit->hp = u_hp(type);
    /* Create fully functional, let other routines set incompleteness. */
    newunit->cp = u_cp(type);
    /* Not in a transport. */
    newunit->transport = NULL;
#if (MAXMTYPES > 2)
    /* This allocates the space for supplies only on the first use. */
    /* Note that the space never needs to be freed. */
    if (newunit->supply == NULL && nummtypes > 0) {
	newunit->supply = (short *) xmalloc(nummtypes * sizeof(short));
    }
#endif
    /* In any case, zero out all the supply values. */
    for_all_material_types(r) newunit->supply[r] = 0;
    /* Need to allocate construction state only if situation requires. */
    if (max_builds(type) > 0 && newunit->tooling == NULL) {
	newunit->tooling = (short *) xmalloc(1 * sizeof(short));
    }
    /* do something about loyalties */
    init_unit_actorstate(newunit);
    init_unit_plan(newunit);
    newunit->occupant = NULL;
    newunit->nexthere = NULL;
    /* Glue this unit into the list of independent units. */
    newunit->next = newunit;
    newunit->prev = newunit;
    insert_unit(indepside->unithead, newunit);
    /* Set more random slots. */
    newunit->prevx = newunit->prevy = -1;
    newunit->hook = lispnil;
    /* Add to the global unit count (is this really necessary?). */
    ++numunits;
    return newunit;
}

/* Alter the actorstate object to be correct for this unit. */

init_unit_actorstate(unit)
Unit *unit;
{
    if (u_acp(unit->type) > 0) {
	/* Might already have an actorstate, don't realloc if so. */
	if (unit->act == NULL) {
	    unit->act = (ActorState *) xmalloc(sizeof(ActorState));
	}
/*	bzero(unit->act, sizeof(ActorState));  */
	unit->act->nextaction.type = A_NONE;
    } else {
	if (unit->act != NULL) free(unit->act);
	unit->act = NULL;
    }
}

/* Every unit that can act needs a plan object, all that cannot should have
   it cleared out. */

init_unit_plan(unit)
Unit *unit;
{
    if (u_acp(unit->type) > 0) {
	/* Might already have a plan, so don't realloc. */
	if (unit->plan == NULL) {
	    unit->plan = (Plan *) xmalloc(sizeof(Plan));
	}
	/* Put the plan into a default state, side will work it up later. */
/*	bzero(unit->plan, sizeof(Plan));  */
	unit->plan->type = UNFORMED_PLAN;
	unit->plan->creationtime = curturn;
	unit->plan->asleep = FALSE;
	unit->plan->reserve = FALSE;
    } else {
	/* Brainless units don't need anything, can free up plan. */
	if (unit->plan != NULL) free(unit->plan);
	unit->plan = NULL;
    }
}

Unit *
designer_create_unit(side, u, s, x, y)
Side *side;
int u, s, x, y;
{
    Unit *newunit;
    Side *side2;

    if ((newunit = create_unit(u)) != NULL) {
	if (s != 0) {
	    side2 = side_n(s);
	    unit_changes_side(newunit, side2, -1, -1);
	}
	init_supply(newunit);
	enter_hex(newunit, x, y);
	see_exact(side, x, y);
	update_cell_display(side, x, y, TRUE);
	update_unit_display(side, newunit, TRUE);
	return newunit;
    } else {
	return NULL;
    }
}

/* Changing the type has many effects. */

change_unit_type(unit, newtype)
Unit *unit;
int newtype;
{
    unit->type = newtype;
    init_unit_actorstate(unit);
    init_unit_plan(unit);
    /* let caller fix hp if desired */
    /* may need to update displays (but after hp changed...) */
}

max_builds(u)
int u;
{
    int u2;

    for_all_unit_types(u2) {
	if (could_make(u, u2)) return 1;
    }
    return 0;
}

/* A unit occupies a hex by adding itself to the list of occupants.
   It will not occupy a transport even if one is at this position
   (other code should have taken care of this case already)
   If something goes wrong, return false.  This routine is heavily used. */

enter_hex(unit, x, y)
Unit *unit;
int x, y;
{
    register int u = unit->type, o;
    register Unit *other = unit_at(x, y);

#ifdef DEBUGGING
    /* Not necessarily an error, but indicative of bugs elsewhere. */
    if (unit->x >= 0 || unit->y >= 0) {
	run_warning("unit %d occupying hex (%d, %d), was at (%d %d)",
		    unit->id, x, y, unit->x, unit->y);
    }
#endif /* DEBUGGING */
    /* Always check this one, but not necessarily fatal. */
    if (!inside_world(x, y)) {
	run_warning("No cell at %d,%d, %s can't enter it",
		    x, y, unit_desig(unit));
	/* Let the unit remain off-world. */
	/* (should mark as inaccessible in some way) */
	return FALSE;
    }
    if (other) {
	if (!can_occupy_cell(unit, x, y)) {
	    Dprintf("unit shouldn't fit here\n");
	    return FALSE;
	}
	unit->nexthere = other;
	/* On top for now, should eventually sort by stack order. */
	set_unit_at(x, y, unit);
    } else {
	unit->nexthere = NULL;
	set_unit_at(x, y, unit);
    }
    /* Set the location slots now. */
    enter_hex_aux(unit, x, y);
    /* Inevitable side-effect of appearing in the new location. */
    all_see_occupy(unit, x, y, TRUE);
    return TRUE;
}

/* Return true if the given unit can fit onto the given hex. */

can_occupy_cell(unit, x, y)
Unit *unit;
int x, y;
{
    int fullness = 0, t = terrain_at(x, y);
    Unit *unit2;

    for_all_stack(x, y, unit2) {
	fullness += ut_size(unit2->type, t);
    }
    return ((fullness + (unit ? ut_size(unit->type, t) : 0)) <= t_capacity(t));
}

type_can_occupy_cell(u, x, y)
int u, x, y;
{
    int fullness = 0, t = terrain_at(x, y);
    Unit *unit2;

    for_all_stack(x, y, unit2) {
	fullness += ut_size(unit2->type, t);
    }
    return ((fullness + ut_size(u, t)) <= t_capacity(t));
}

/* Recursive helper to update everybody's position.  This should be one of
   two routines that modify actual unit positions (leaving is the other). */

/* *Every* occupant will increment viewing coverage - strains realism, */
/* but prevents strange bugs. */
/* (should be occupants that are not packed up) */

enter_hex_aux(unit, x, y)
Unit *unit;
int x, y;
{
    register Unit *occ;

    /* Actually set the unit position. */
    unit->x = x;  unit->y = y;
    /* Increment viewing coverage. */
    if (unit->transport == NULL
	|| uu_occ_can_see(unit->type, unit->transport->type)) {
	cover_area(unit, x, y, 1);
    }
#if 0
    /* Increment special viewing coverage. */
    if (unit->transport == NULL
	|| uu_sp_occ_can_see(unit->type, unit->transport->type)) {
	sp_cover_area(unit, x, y, 1);
    }
#endif
    /* Do for all the occupants too, recursively. */
    for_all_occupants(unit, occ) enter_hex_aux(occ, x, y);
}

/* Decide whether the given unit can actually be in the given transport. */
/* (this still needs to account for multipart units) */

can_occupy(unit, transport)
Unit *transport, *unit;
{
    return can_carry(transport, unit);
}

can_carry(transport, unit)
Unit *transport, *unit;
{
    int u = unit->type, u2 = transport->type;
    int numthistype = 0, occvolume = 0;
    Unit *occ;

    /* Why is this here? */
    if (transport == unit) return FALSE;
    if (unit->occupant != NULL && !uu_occ_can_have_occs(u, u2)) return FALSE;
    /* Compute the transport's fullness. */
    for_all_occupants(transport, occ) {
	if (occ->type == u) ++numthistype;
	occvolume += uu_size(occ->type, u2);
    }
    /* Can carry if dedicated space available. */
    if (numthistype + 1 <= uu_capacity_x(u2, u)) return TRUE;
    /* Can carry if general unit hold has room. */
    return (occvolume + uu_size(u, u2) <= u_capacity(u2));
}

type_can_occupy(u, transport)
int u;
Unit *transport;
{
    int u2 = transport->type;
    int numthistype = 0, occvolume = 0;
    Unit *occ;

    /* Compute the transport's fullness. */
    for_all_occupants(transport, occ) {
	if (occ->type == u) ++numthistype;
	occvolume += uu_size(occ->type, u2);
    }
    /* Can carry if dedicated space available. */
    if (numthistype + 1 <= uu_capacity_x(u2, u)) return TRUE;
    /* Can carry if general unit hold has room. */
    return (occvolume + uu_size(u, u2) <= u_capacity(u2));
}

can_carry_type(transport, u)
Unit *transport;
int u;
{
    return type_can_occupy(u, transport);
}

/* Units become passengers by linking into the transport's passenger list. */

/* (define what else happens) */
/* (should change form to whatever fits in transport?) */

void
enter_transport(unit, transport)
Unit *unit, *transport;
{
    /* Link things together. */
    unit->nexthere = transport->occupant;
    transport->occupant = unit;
    unit->transport = transport;
    /* Move things to their "correct" places. */
    sort_occupants_by_speed(transport);
    /* Set the passenger's coords to match the transport's. */
    enter_hex_aux(unit, transport->x, transport->y);
    /* Others might be watching. */
    all_see_occupy(unit, transport->x, transport->y, FALSE);
}

/* Unit departs from a hex by zeroing out pointer if in hex or by being
   removed from the list of transport occupants. */

/* Dead units (hp = 0) may be run through here, so don't error out. */

void
leave_hex(unit)
Unit *unit;
{
    int ux = unit->x, uy = unit->y;
    Unit *other;

    if (ux < 0 || uy < 0) {
	/* Sometimes called twice */
    } else if (unit->transport != NULL) {
	leave_transport(unit);
	leave_hex_aux(unit);
	all_see_leave(unit, ux, uy, FALSE);
	/* not all_see_hex here because can't see inside transports */
    } else {
	/* Unsplice ourselves from the list of units in this hex. */
	if (unit == unit_at(ux, uy)) {
	    set_unit_at(ux, uy, unit->nexthere);
	} else {
	    for_all_stack(ux, uy, other) {
		if (unit == other->nexthere) {
		    other->nexthere = other->nexthere->nexthere;
		    break;
		}
	   } 
	}
	/* Bash this now-spurious link. */
	unit->nexthere = NULL;
	/* Now bash the coords. */
	leave_hex_aux(unit);
	/* Now let everybody observe that the unit is gone. */
	all_see_leave(unit, ux, uy, TRUE);
    }
}

/* When leaving, remove view coverage, record old position, and then
   trash the old coordinates just in case.  Catches many bugs.  Do
   this for all the occupants as well. */

leave_hex_aux(unit)
Unit *unit;
{
    register Unit *occ;

    /* Decrement viewing coverage around our about-to-be-old location. */
    if (unit->transport == NULL
	|| uu_occ_can_see(unit->type, unit->transport->type)) {
	cover_area(unit, unit->x, unit->y, -1);
    }
#if 0
    /* Decrement special viewing coverage. */
    if (unit->transport == NULL
	|| uu_sp_occ_can_see(unit->type, unit->transport->type)) {
	sp_cover_area(unit, unit->x, unit->y, -1);
    }
#endif
    /* Stash the old coords. */
    unit->prevx = unit->x; unit->prevy = unit->y;
    /* Set to a recognizable value. */
    unit->x = -1;  unit->y = -1;
    /* Make all the occupants leave too. */
    for_all_occupants(unit, occ) leave_hex_aux(occ);
}

/* Disembarking unlinks from the list of passengers only, leaves the unit
   hanging in limbo, so should have it occupy something immediately. */

void
leave_transport(unit)
Unit *unit;
{
    Unit *transport = unit->transport, *occ;

    if (unit == transport->occupant) {
	transport->occupant = unit->nexthere;
    } else {
	for_all_occupants(transport, occ) {
	    if (unit == occ->nexthere) {
		occ->nexthere = occ->nexthere->nexthere;
		break;
	    }
	}
    }
    /* Bash the now-spurious link. */
    unit->transport = NULL;
}

/* Handle the general situation of a unit changing allegiance from one side
   to another.  This is a common internal routine, so no messages here. */

unit_changes_side(unit, newside, reason1, reason2)
Unit *unit;
Side *newside;
int reason1, reason2;
{
    Side *oldside = unit->side;
    Unit *occ;

    if (oldside == newside) return;
    if (!unit_allowed_on_side(unit, newside)) {
	if (reason1 == CAPTURE) kill_unit(unit, -1);
	return;
    }
    set_unit_side(unit, newside);
    /* Give captured units some additional moves maybe. */
#if 0
    if (reason1 == CAPTURE)
      unit->movesleft = (period_capturemoves() ? compute_move(unit) : 0);
#endif
    /* Remove unit from former side's inventory and counts. */
    if (oldside != NULL) {
/*	oldside->units[unit->type]--;  */
/*	if (producing(unit)) oldside->unitsbuilding[unit->product]--;  */
/*	if (reason2 >= 0) side_balance(oldside, unit->type, reason2)++; */
	update_unit_display(oldside, unit);
	/* for machine players */
/*	if (reason2 == PRISONER) register_loss_to_area(oldside, unit, 2);  */
    }
    /* Add unit to new side's inventory and counts. */
    if (newside != NULL) {
/*	newside->units[unit->type]++;  */
	/*	    if (producing(unit)) newside->unitsbuilding[unit->product]++;  */
	/*	    if (reason1 >= 0) side_balance(newside, unit->type, reason1)++; */
    }
    /* (Should this be switchable maybe?) */
    for_all_occupants(unit, occ) {
	unit_changes_side(occ, newside, reason1, reason2);
    }
    /* Adjust view coverage.  The sequencing here is to make sure that no
       viewing coverage gets left on or off inadvertantly. */
    if (alive(unit)) {
	cover_area(unit, unit->x, unit->y, -1);
	/* Actually set the side slot of the unit here. */
	set_unit_side(unit, newside);
	cover_area(unit, unit->x, unit->y, 1);
    }
    /* Reflect the changeover in any appropriate displays. */
    if (oldside != NULL) {
	update_unit_display(oldside, unit); /* a second update? */
    }
    if (newside != NULL) {
	update_unit_display(newside, unit);
    }
}

/* This is a general test as to whether the given unit can be
   on the given side. */

unit_allowed_on_side(unit, side)
Unit *unit;
Side *side;
{
    if (unit == NULL) return FALSE;
    /* (allow for tests on individual units also?) */
    return type_allowed_on_side(unit->type, side);
}

test_class_membership(leaf)
Obj *leaf;
{
    char *sclass;

    if (stringp(leaf)) {
	sclass = c_string(leaf);
	if (tmpside != NULL) {
	    if (empty_string(tmpside->sideclass)) return FALSE;
	    return (strcmp(sclass, tmpside->sideclass) == 0);
	} else {
	    return (strcmp(sclass, "independent") == 0);
	}
    } else {
	init_warning("testing side against garbled class expression");
	/* Be permissive if continued. */
	return TRUE;
    }
}

eval_boolean_expression(expr, fn, dflt)
Obj *expr;
int (*fn)(), dflt;
{
    char *opname;

    if (expr == lispnil) {
	return dflt;
    } else if (consp(expr) && symbolp(car(expr))) {
	opname = c_string(car(expr));
	if (strcmp(opname, "and") == 0) {
	    return (eval_boolean_expression(cadr(expr), fn, dflt)
		    && eval_boolean_expression(car(cddr(expr)), fn, dflt));
	} else if (strcmp(opname, "or") == 0) {
	    return (eval_boolean_expression(cadr(expr), fn, dflt)
		    || eval_boolean_expression(car(cddr(expr)), fn, dflt));
	} else if (strcmp(opname, "not") == 0) {
	    return !eval_boolean_expression(cadr(expr), fn, dflt);
	} else {
	    return (*fn)(expr);
	}
    } else {
	return (*fn)(expr);
    }
}

type_allowed_on_side(u, side)
int u;
Side *side;
{
    Obj *sclass = u_possible_sides(u);

    /* (could optimize by precomputing bit vectors) */
    tmputype = u;
    tmpside = side;
    return eval_boolean_expression(sclass, test_class_membership, TRUE);
}

/* Put the given unit on the given side, without all the fancy effects.
   Important to handle independents, because this gets called during init.
   This is the only way that a unit's side may be altered. */

set_unit_side(unit, side)
Unit *unit;
Side *side;
{
    if (unit->side != side) {
	if (!unit_allowed_on_side(unit, side)) return FALSE;
	unit->side = side;
	/* Make sure this unit is off anybody else's list. */
	delete_unit(unit);
	insert_unit((side ? side->unithead : indepside->unithead), unit);
    }
    return TRUE;
}


/* Remove a unit from play.  This is different from making it available for
   reallocation - only the unit flusher can do that.  We remove all the
   passengers too, recursively.  Sometimes units are "killed twice", so
   be sure not to run all this twice.  Also count up occupant deaths, being
   sure not to count the unit itself as an occupant. */

kill_unit(unit, reason)
Unit *unit;
int reason;
{
    int u = unit->type, u2, selfdied = FALSE;

    if (alive(unit)) {
	if (unit->side && unit->side->selfunit == unit) selfdied = TRUE;
	for_all_unit_types(u2) occdeath[u2] = 0;
	leave_hex(unit);
	kill_unit_aux(unit, reason);
	if (selfdied) {
	    if (u_self_resurrects(u)) {
		/* find and designate a new self unit */
		return;
	    } else {
		/* Make sure this doesn't have serious consequences? */
		/* (actually, side might have won??  should just take out
		   of game permanently) */
		side_loses(unit->side, NULL);
		/* (can't do all consequences just yet?) */
	    }
	}
	occdeath[u]--;
	recent_dead_flushed = FALSE;
    }
}

/* Trash it now - occupant doesn't need to leave_hex.  Also record the
   event, and update the apropriate display.  The unit here should
   be known to be alive. */

kill_unit_aux(unit, reason)
Unit *unit;
int reason;
{
    int u = unit->type;
    Unit *occ;
    
    unit->hp = 0;
    occdeath[u]++;
/*    dispose_of_plan(unit);  for now... */
    if (reason >= 0) {
	record_unit_loss(unit, reason);
	if (unit->side != NULL) {
/*	    side_balance(unit->side, u, reason)++; */
/*	    unit->side->units[u]--;  */
	    /*	if (producing(unit)) unit->side->unitsbuilding[unit->product]--;  */
	    react_to_unit_loss(unit->side, unit);
	    update_unit_display(unit->side, unit);
	}
    }
    /* Kill all the occupants too. */
    for_all_occupants(unit, occ) if (alive(occ)) kill_unit_aux(occ, reason);
}

/* Give strategists a chance to react. */

react_to_unit_loss(side, unit)
Side *side;
Unit *unit;
{
    if (side_has_ai(side)) {
	ai_react_to_unit_loss(side, unit);
    }
}

/* Do something about the plan object, possibly passing to a deputy unit,
   or maybe just destroying it entirely. */

dispose_of_plan(unit)
Unit *unit;
{
    free_plan(unit->plan);
    unit->plan = NULL;
}

/* Get rid of all dead units at once.  It's important to get rid of all
   of the dead units, so they don't come back and haunt us.
   (This routine is basically a garbage collector, and should not be called
   during a unit list traversal.) The process starts by finding the first
   live unit, making it the head, then linking around all in the middle.
   Dead units stay on the dead unit list for each side until that side has had
   a chance to move.  Then they are finally flushed in a permanent fashion. */ 

void
flush_dead_units()
{
#ifdef REUSE_DEAD
    Unit *unit, *prevunit;
    Side *side;

    /* This never ends up flushing independent units, no big loss. */
    if (!recent_dead_flushed) {
	recent_dead_flushed = TRUE;
	for_all_units(side, unit) {
	    if (!alive(unit)) {
		prevunit = unit->prev;
		delete_unit(unit);
		put_unit_on_dead_list(unit);
		unit = prevunit;
	    }
	}
    }
#endif
}

/* Put dead units on sides dead unit lists. */

put_unit_on_dead_list(unit)
Unit *unit;
{
    if (unit->side == NULL) {
	flush_one_unit(unit);
    } else {
	unit->next = unit->side->deadunits;
	unit->side->deadunits = unit;
    }
    --numunits;
}

/* Clean up dead units and put them back on free list. */

flush_side_dead(side)
Side *side;
{
    Unit *unit, *next;
  
    next = side->deadunits;
    while (next != NULL) {
	unit = next;
	next = unit->next;
	flush_one_unit(unit);
    }
    side->deadunits = NULL;
}

/* Keep it clean - hit all links to other places.  Some might not be
   strictly necessary, but this is not an area to take chances with. */

flush_one_unit(unit)
Unit *unit;
{
    unit->id = -1;
    unit->occupant = NULL;
    unit->transport = NULL;
    unit->nexthere = NULL;
    /* Add it on the list of available units. */
    unit->next = freeunits;
    freeunits = unit;
}

/* Do multiple passes of bubble sort.  This is intended to improve locality
   among unit positions and reduce the amount of scrolling around.  It is
   not required for correct behavior anywhere (I think).
   Data is generally coherent, so bubble sort not too bad if we allow
   early termination when everything is already in order.  */

/* If slowness objectionable, replace with something clever, but be
   sure that average performance in real games is what's being improved. */

void
sort_units()
{
    int flips;
    int passes = 0;
    register Unit *unit, *nextunit;
    Side *side;

    for_all_sides(side) {
	passes = 0;
	flips = TRUE;
	while (flips) {
	    flips = FALSE;
	    for_all_side_units(side,unit) {
		if (unit->next != side->unithead
		    && !in_order(unit, unit->next)) {
		    flips = TRUE;
		    /* Reorder the units by fiddling with their links. */
		    nextunit = unit->next;
		    unit->prev->next = nextunit;
		    nextunit->next->prev = unit;
		    nextunit->prev = unit->prev;
		    unit->next = nextunit->next;
		    nextunit->next = unit;
		    unit->prev = nextunit;
		}
		++passes;
	    }
	}
    }
    Dprintf("Sorting passes = %d\n", passes);
}

in_order(unit1, unit2)
Unit *unit1, *unit2;
{
    int d1, d2;

    if (unit1->type < unit2->type) return TRUE;
    if (unit1->type > unit2->type) return FALSE;
    /* (should sort named units alphabetically) */
    /* Last gasp - ids impose a total ordering. */
    return (unit1->id < unit2->id);
}

/* A unit runs low on supplies at the halfway point, but only worries about */
/* the essential items.  Formula is the same no matter how/if occupants eat */
/* transports' supplies. */

low_supplies(unit)
Unit *unit;
{
    int u = unit->type, r;

    for_all_material_types(r) {
	if (((um_consume(u, r) > 0) || (um_tomove(u, r) > 0)) &&
	    /* Really should check that the transport is adequate for */
	    /* supplying the fuel */ 
	    (unit->transport == NULL)) {
	    if (2 * unit->supply[r] <= um_storage(u, r)) return TRUE;
	}
    }
    return FALSE;
}

/* Useful for the machine player to know how long it can move this
   piece before it should go home.  Assumes can't replenish from
   terrain.  Result may be negative, in which case it's time to go! */

moves_till_low_supplies(unit)
Unit *unit;
{
    int u = unit->type, r, moves = 12345, tmp;

    for_all_material_types(r) {
	if ((um_tomove(u, r) > 0)) {
	    tmp = (unit->supply[r] - um_storage(u, r) / 2) / um_tomove(u, r);
	    moves = min(moves, tmp);
	}
    }
    return moves;
}

/* Short unreadable but greppable listing of unit. */

char *
unit_desig(unit)
Unit *unit;
{
    if (unit != NULL) {
	if (unit->id == -1) {
	    sprintf(shortbuf, "s%d head", side_number(unit->side));
	} else if (is_unit_type(unit->type)) {
	    sprintf(shortbuf, "s%d %s %d (%d,%d",
		    side_number(unit->side), utype_name_n(unit->type, 1),
		    unit->id, unit->x, unit->y);
	    if (unit->z != 0) {
		sprintf(shortbuf+strlen(shortbuf), ",%d", unit->z);
	    }
	    if (((int) unit->transport) > 1000 /* != NULL */) {
		sprintf(shortbuf+strlen(shortbuf), ",in%d",
			unit->transport->id);
	    }
	    strcat(shortbuf, ")");  /* close out the unit coords */
	} else {
	    sprintf(shortbuf, "!garbage unit!");
	}
    } else {
	sprintf(shortbuf, "no unit");
    }
    return shortbuf;
}

/* Come up with a unit type name that fits in the given space. */

char utypenamen[100];

char *
utype_name_n(u, n)
int u, n;
{
    if (n <= 0) {
	/* Allowing infinite space. */
	return u_type_name(u);
    } else if (!empty_string(u_short_name(u))) {
	if (strlen(u_short_name(u)) <= n) {
	    return u_short_name(u);
	} else {
	    /* Copy what will fit. */
	    strncpy(utypenamen, u_short_name(u), n);
	    return utypenamen;
	}
    } else {
	/* Copy what will fit. */
	strncpy(utypenamen, u_type_name(u), n);
	return utypenamen;
    }
}

char *
actorstate_desig(as)
ActorState *as;
{
    if (as != NULL) {
	sprintf(actorstatebuf, "(acp %d/%d) (mp %d/%d) %s",
		as->acp, as->initacp, as->mp, as->initmp,
		action_desig(as->nextaction));
	return actorstatebuf;
    } else {
	return "no act";
    }
}

/* General-purpose routine to take an array of anonymous unit types and */
/* summarize what's in it, using numbers and unit chars. */

char *
summarize_units(buf, ucnts)
char *buf;
int *ucnts;
{
    char tmp[BUFSIZE];  /* should be bigger? */
    int u;

    sprintf(buf, "");
    for_all_unit_types(u) {
	if (ucnts[u] > 0) {
	    sprintf(tmp, " %d %s", ucnts[u], utype_name_n(u, 3));
	    strcat(buf, tmp);
	}
    }
    return buf;
}

/* Search for a unit with the given id number. */

/* If this is used a lot, it should be sped up. */

Unit *
find_unit(n)
int n;
{
    Unit *unit;
    Side *loopside;

    for_all_units(loopside, unit)
      if (alive(unit) && unit->id == n) return unit;
    return NULL;
}

/* Given a name, find the type. */

find_unit_name(str)
char *str;
{
    int u;

    for_all_unit_types(u)
      if (strcmp(str, u_short_name(u)) == 0
	  || strcmp(str, u_type_name(u)) == 0) 
	return u;
    return NONUTYPE;
}

/* Reverse the order of the unit list.  This is needed since the
   loaded units from game modules are put into the list in reverse order. */

reverse_unit_order()
{
    Unit *next, *unit;
    Side *side;

    for_all_sides(side) {
	next = side->unithead->next;
	do {
	    unit = next;
	    next = unit->next;
	    unit->next = unit->prev;
	    unit->prev = next;
	} while (unit != side->unithead);
    }
}    

sort_occupants_by_speed(trans)
Unit *trans;
{
    int flips = TRUE;
    Unit *prevunit, *unit, *nextunit;

    while (flips) {
	flips = FALSE;
	prevunit = NULL;
	unit = trans->occupant;
	while (unit != NULL) {
	    if (unit->nexthere != NULL &&
		u_speed(unit->type) < u_speed(unit->nexthere->type)) {
		flips = TRUE;
		nextunit = unit->nexthere;
		if (prevunit != NULL) 
		  prevunit->nexthere = nextunit;
		else
		  trans->occupant = nextunit;
		unit->nexthere = nextunit->nexthere;
		nextunit->nexthere = unit;
		prevunit = nextunit;
	    } else {
		prevunit = unit;
		unit = unit->nexthere;
	    }
	}
    }
}

/* Return the first unit on the list. */

Unit *
first_unit (side)
Side *side;
{
    return side->unithead;
}

/* Insert the given unit after the other given unit. */

void
insert_unit(unithead, unit)
Unit *unithead, *unit;
{
    unit->next = unithead->next;
    unit->prev = unithead;
    unithead->next->prev = unit;
    unithead->next = unit;
}

/* Delete the unit from its list. */

void
delete_unit(unit)
Unit *unit;
{
    unit->next->prev = unit->prev;
    unit->prev->next = unit->next;
}

num_occupants(unit)
Unit *unit;
{
    int num = 0;
    Unit *occ;

    for_all_occupants(unit, occ) {
	num += 1;
    }
    return num;
}

num_units_at(x, y)
int x, y;
{
    int num = 0;
    Unit *unit;

    x = wrapx(x);
    if (!in_area(x, y)) {
	run_warning("num_units_at %d,%d??", x, y);
	return 0;
    }
    for_all_stack(x, y, unit) {
	num += 1;
    }
    return num;
}
